/*
  Copyright (c) 2005, The Musepack Development Team
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

  * Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.

  * Neither the name of the The Musepack Development Team nor the
  names of its contributors may be used to endorse or promote
  products derived from this software without specific prior
  written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/// \file streaminfo.c
/// Implementation of streaminfo reading functions.

#include <mpcdec/mpcdec.h>
#include <mpcdec/streaminfo.h>
#include <stdio.h>
#include "internal.h"

mpc_uint32_t mpc_bits_read(mpc_bits_reader * r, const unsigned int nb_bits);
unsigned int mpc_bits_get_size(mpc_bits_reader * r, mpc_uint64_t * p_size);

static const char na[] = "n.a.";
static const char *versionNames[] = {
    na, "'Unstable/Experimental'", na, na, na, "'quality 0'", "'quality 1'",
    "'Telephone'", "'Thumb'", "'Radio'", "'Standard'", "'Extreme'", "'Insane'",
    "'BrainDead'", "'quality 9'", "'quality 10'"
};
static const mpc_int32_t samplefreqs[4] = { 44100, 48000, 37800, 32000 };

static const char *
mpc_get_version_string(mpc_uint32_t profile) // profile is 0...15, where 7...13 is used
{
    return profile >= sizeof versionNames / sizeof *versionNames ? na : versionNames[profile];
}

static void
mpc_get_encoder_string(mpc_streaminfo* si)
{
	if (si->encoder_version == 0) {
		sprintf(si->encoder, "Buschmann 1.7.0...9, Klemm 0.90...1.05");
	} else {
		switch (si->encoder_version % 10) {
			case 0:
				sprintf(si->encoder, "Release %u.%u", si->encoder_version / 100,
						si->encoder_version / 10 % 10);
				break;
			case 2: case 4: case 6: case 8:
				sprintf(si->encoder, "Beta %u.%02u", si->encoder_version / 100,
						si->encoder_version % 100);
				break;
			default:
				sprintf(si->encoder, "--Alpha-- %u.%02u",
						si->encoder_version / 100, si->encoder_version % 100);
				break;
		}
	}
}

/// Reads streaminfo from SV7 header.
mpc_status
streaminfo_read_header_sv7(mpc_streaminfo* si, mpc_bits_reader * r)
{
    mpc_uint16_t Estimatedpeak_title = 0;

	si->bitrate            = 0;
	si->block_size         = 1;
	si->frames             = mpc_bits_read(r, 32);
	si->is                 = mpc_bits_read(r, 1); // should be 0
	si->ms                 = mpc_bits_read(r, 1);
	si->max_band           = mpc_bits_read(r, 6);
	si->profile            = mpc_bits_read(r, 4);
    si->profile_name       = mpc_get_version_string(si->profile);
	mpc_bits_read(r, 2); // Link ?
	si->sample_freq        = samplefreqs[mpc_bits_read(r, 2)];
	Estimatedpeak_title    = (mpc_uint16_t) mpc_bits_read(r, 16);   // read the ReplayGain data
	si->gain_title         = (mpc_uint16_t) mpc_bits_read(r, 16);
	si->peak_title         = (mpc_uint16_t) mpc_bits_read(r, 16);
	si->gain_album         = (mpc_uint16_t) mpc_bits_read(r, 16);
	si->peak_album         = (mpc_uint16_t) mpc_bits_read(r, 16);
	si->is_true_gapless    = mpc_bits_read(r, 1); // true gapless: used?
	si->last_frame_samples = mpc_bits_read(r, 11); // true gapless: valid samples for last frame
	si->fast_seek          = mpc_bits_read(r, 1); // fast seeking
	mpc_bits_read(r, 19); // unused
	si->encoder_version    = mpc_bits_read(r, 8);
    si->channels           = 2;

	mpc_get_encoder_string(si);

	// estimation, exact value needs too much time
	si->pcm_samples     = 1152 * si->frames - 576;
	si->average_bitrate = 0;
	if (si->pcm_samples > 0)
		si->average_bitrate = (si->tag_offset  - si->header_position) * 8.0
				*  si->sample_freq / si->pcm_samples;

    return MPC_STATUS_OK;
}

/// Reads streaminfo from SV8 header.
mpc_status
streaminfo_read_header_sv8(mpc_streaminfo* si, const mpc_bits_reader * r_in)
{
	mpc_uint32_t CRC;
	mpc_uint64_t sampleCount = 0;
	mpc_bits_reader r = *r_in;

	// FIXME : add CRC check
	CRC = mpc_bits_read(&r, 32);
	si->stream_version = mpc_bits_read(&r, 8);
	mpc_bits_get_size(&r, &sampleCount);
	si->frames = (sampleCount + MPC_FRAME_LENGTH - 1) / MPC_FRAME_LENGTH;
	si->is_true_gapless = 1;
	si->last_frame_samples = sampleCount - (si->frames - 1) * MPC_FRAME_LENGTH;
	si->sample_freq = samplefreqs[mpc_bits_read(&r, 4)];
	si->channels = mpc_bits_read(&r, 4) + 1;
	si->max_band = mpc_bits_read(&r, 5) + 1;
	// FIXME : this flag will be removed, must be 0
	si->is = mpc_bits_read(&r, 1);
	si->ms = mpc_bits_read(&r, 1);

	si->bitrate            = 0;
	si->block_size         = 1;

	si->pcm_samples     = sampleCount;
	si->average_bitrate = 0;
	if (si->pcm_samples > 0)
		si->average_bitrate = (si->tag_offset - si->header_position) * 8.0
				*  si->sample_freq / si->pcm_samples;

	return MPC_STATUS_OK;
}

double
mpc_streaminfo_get_length(mpc_streaminfo * si)
{
    return (double) mpc_streaminfo_get_length_samples(si) / si->sample_freq;
}

mpc_int64_t
mpc_streaminfo_get_length_samples(mpc_streaminfo * si)
{
    mpc_int64_t samples = (mpc_int64_t) si->frames * MPC_FRAME_LENGTH;
    if (si->is_true_gapless)
        samples -= (MPC_FRAME_LENGTH - si->last_frame_samples);
    else
        samples -= MPC_DECODER_SYNTH_DELAY;
    return samples;
}
